@@ -2,13 +2,7 @@
 
 import Control.Applicative
 import Control.Monad.Logic
-#if !MIN_VERSION_base(4,8,0)
-import Data.Monoid (Monoid (..))
-#endif
-#if MIN_VERSION_base(4,9,0)
 import Data.Semigroup (Semigroup (..))
-#endif
-
 
 parents :: [ (String, String) ]
 parents = [ ("Sarah",  "John")
