@@ -1,12 +1,12 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Main where
 
 import           Test.Tasty
 import           Test.Tasty.HUnit
 
-import           Control.Arrow ( left )
 import           Control.Concurrent ( threadDelay )
 import           Control.Concurrent.Async ( race )
 import           Control.Exception
@@ -16,20 +16,23 @@ import           Control.Monad.Logic
 import           Control.Monad.Reader
 import qualified Control.Monad.State.Lazy as SL
 import qualified Control.Monad.State.Strict as SS
+import           Data.List (uncons)
 import           Data.Maybe
 
-#if MIN_VERSION_base(4,9,0) && !MIN_VERSION_base(4,11,0)
-import           Data.Semigroup (Semigroup (..))
+#if MIN_VERSION_base(4,17,0)
+import GHC.IsList (IsList(..))
+#else
+import GHC.Exts (IsList(..))
 #endif
 
--- required by base < 4.9 OR CPS Writer test
-#if !MIN_VERSION_base(4,9,0) || MIN_VERSION_mtl(2,3,0)
-import           Data.Monoid
+#if !MIN_VERSION_base(4,11,0)
+import           Data.Semigroup (Semigroup (..))
 #endif
 
 #if MIN_VERSION_mtl(2,3,0)
 import qualified Control.Monad.Writer.CPS as CpsW (WriterT, execWriterT, tell)
 import qualified Control.Monad.Trans.Writer.CPS as CpsW (runWriterT)
+import           Data.Monoid
 #endif
 
 monadReader1 :: Assertion
@@ -63,13 +66,9 @@ evalWriterT :: (Monad m, Monoid w) => CpsW.WriterT w m a -> m a
 evalWriterT = fmap fst . CpsW.runWriterT
 #endif
 
-#if MIN_VERSION_base(4,8,0)
 nats = pure 0 `mplus` ((1 +) <$> nats)
-#else
-nats = return 0 `mplus` liftM (1 +) nats
-#endif
 
-odds = return 1 `mplus` liftM (2+) odds
+odds = return 1 `mplus` fmap (2+) odds
 
 oddsOrTwoUnfair = odds `mplus` return 2
 oddsOrTwoFair   = odds `interleave` return 2
@@ -87,9 +86,7 @@ yieldWords = go
 
 main :: IO ()
 main = defaultMain $
-#if __GLASGOW_HASKELL__ >= 702
   localOption (mkTimeout 3000000) $  -- 3 second deadman timeout
-#endif
   testGroup "All"
   [ testGroup "Monad Reader + env"
     [ testCase "Monad Reader 1" monadReader1
@@ -104,7 +101,7 @@ main = defaultMain $
       testCase "runIdentity all"  $ [0..4] @=? (take 5 $ runIdentity $ observeAllT nats)
     , testCase "runIdentity many" $ [0..4] @=? (runIdentity $ observeManyT 5 nats)
     , testCase "observeAll"       $ [0..4] @=? (take 5 $ observeAll nats)
-    , testCase "observeMany"      $ [0..4] @=? (observeMany 5 nats)
+    , testCase "observeMany"      $ [0..4] @=? observeMany 5 nats
 
     -- Ensure LogicT can be run over other base monads other than
     -- List.  Some are productive (Reader) and some are non-productive
@@ -115,7 +112,7 @@ main = defaultMain $
 
     , testCase "observeManyT can be used with Either" $
       (Right [0..4] :: Either Char [Integer]) @=?
-      (observeManyT 5 nats)
+      observeManyT 5 nats
     ]
 
   --------------------------------------------------
@@ -124,14 +121,14 @@ main = defaultMain $
     [
       testCase "runLogicT multi" $ ["Hello world !"] @=?
       let conc w o = fmap ((w `mappend` " ") `mappend`) o in
-      (runLogicT (yieldWords ["Hello", "world"]) conc (return "!"))
+      runLogicT (yieldWords ["Hello", "world"]) conc (return "!")
 
     , testCase "runLogicT none" $ ["!"] @=?
       let conc w o = fmap ((w `mappend` " ") `mappend`) o in
-      (runLogicT (yieldWords []) conc (return "!"))
+      runLogicT (yieldWords []) conc (return "!")
 
     , testCase "runLogicT first" $ ["Hello"] @=?
-      (runLogicT (yieldWords ["Hello", "world"]) (\w -> const $ return w) (return "!"))
+      runLogicT (yieldWords ["Hello", "world"]) (\w -> const $ return w) (return "!")
 
     , testCase "runLogic multi" $ 20 @=? runLogic odds5down (+) 11
     , testCase "runLogic none"  $ 11 @=? runLogic mzero (+) (11 :: Integer)
@@ -144,6 +141,12 @@ main = defaultMain $
 
     , testCase "observeMany multi" $ [5,3] @=? observeMany 2 odds5down
     , testCase "observeMany none" $ ([] :: [Integer]) @=? observeMany 2 mzero
+
+    , testCase "(>>-) Logic" $ do
+        let sample = fromList [1, 2, 3] :: Logic Integer
+        (sample >>- const (mempty :: Logic Integer)) @?= mempty
+        (sample >>- (\x -> fmap (+ x) (fromList [100, 200, 300]))) @?= fromList [101,102,201,103,301,202,203,302,303]
+        (sample >>- (\x -> if odd x then fmap (+ x) (fromList [100, 200, 300]) else mempty)) @?= fromList [101,103,201,203,301,303]
     ]
 
   --------------------------------------------------
@@ -172,7 +175,7 @@ main = defaultMain $
         , testCase "msplit mzero :: LogicT" $
           let z :: LogicT [] String
               z = mzero
-          in assertBool "LogicT" $ null $ catMaybes $ concat $ observeAllT (msplit z)
+          in assertBool "LogicT" $ all (null . catMaybes) $ observeAllT (msplit z)
         , testCase "msplit mzero :: strict StateT" $
           let z :: SS.StateT Int [] String
               z = mzero
@@ -192,6 +195,11 @@ main = defaultMain $
             extract (msplit op) @?= [Just 1]
             extract (msplit op >>= (\(Just (_,nxt)) -> msplit nxt)) @?= [Just 2]
 
+        , testCase "(>>-) []" $ do
+            (sample >>- const ([] :: [Integer])) @?= []
+            (sample >>- (\x -> fmap (+ x) [100, 200, 300])) @?= [101,102,201,103,301,202,203,302,303]
+            (sample >>- (\x -> if odd x then fmap (+ x) [100, 200, 300] else [])) @?= [101,103,201,203,301,303]
+
         , testCase "msplit ReaderT" $ do
             let op = ask
                 extract = fmap fst . catMaybes . flip runReaderT sample
@@ -210,20 +218,20 @@ main = defaultMain $
         , testCase "msplit LogicT" $ do
             let op :: LogicT [] Integer
                 op = foldr (mplus . return) mzero sample
-                extract = fmap fst . catMaybes . concat . observeAllT
+                extract = fmap fst . concatMap catMaybes . observeAllT
             extract (msplit op) @?= [1]
             extract (msplit op >>= (\(Just (_,nxt)) -> msplit nxt)) @?= [2]
 
         , testCase "msplit strict StateT" $ do
             let op :: SS.StateT Integer [] Integer
-                op = (SS.modify (+1) >> SS.get `mplus` op)
+                op = SS.modify (+1) >> SS.get `mplus` op
                 extract = fmap fst . catMaybes . flip SS.evalStateT 0
             extract (msplit op) @?= [1]
             extract (msplit op >>= \(Just (_,nxt)) -> msplit nxt) @?= [2]
 
         , testCase "msplit lazy StateT" $ do
             let op :: SL.StateT Integer [] Integer
-                op = (SL.modify (+1) >> SL.get `mplus` op)
+                op = SL.modify (+1) >> SL.get `mplus` op
                 extract = fmap fst . catMaybes . flip SL.evalStateT 0
             extract (msplit op) @?= [1]
             extract (msplit op >>= \(Just (_,nxt)) -> msplit nxt) @?= [2]
@@ -263,26 +271,26 @@ main = defaultMain $
                   in oddsOrTwoLFair)
 
       , testCase "fair disjunction :: ReaderT" $ [1,2,3,5] @=?
-        (take 4 $ runReaderT (let oddsR = return 1 `mplus` liftM (2+) oddsR
+        (take 4 $ runReaderT (let oddsR = return 1 `mplus` fmap (2+) oddsR
                               in oddsR `interleave` return (2 :: Integer)) "go")
 
 #if MIN_VERSION_mtl(2,3,0)
       , testCase "fair disjunction :: CPS WriterT" $ [1,2,3,5] @=?
         (take 4 $ evalWriterT (let oddsW :: CpsW.WriterT [Char] [] Integer
-                                   oddsW = return 1 `mplus` liftM (2+) oddsW
+                                   oddsW = return 1 `mplus` fmap (2+) oddsW
                                 in oddsW `interleave` return (2 :: Integer)))
 #endif
 
       , testCase "fair disjunction :: strict StateT" $ [1,2,3,5] @=?
-        (take 4 $ SS.evalStateT (let oddsS = return 1 `mplus` liftM (2+) oddsS
+        (take 4 $ SS.evalStateT (let oddsS = return 1 `mplus` fmap (2+) oddsS
                                   in oddsS `interleave` return (2 :: Integer)) "go")
 
       , testCase "fair disjunction :: lazy StateT" $ [1,2,3,5] @=?
-        (take 4 $ SL.evalStateT (let oddsS = return 1 `mplus` liftM (2+) oddsS
+        (take 4 $ SL.evalStateT (let oddsS = return 1 `mplus` fmap (2+) oddsS
                                   in oddsS `interleave` return (2 :: Integer)) "go")
       ]
 
-    , testGroup "fair conjunction" $
+    , testGroup "fair conjunction"
       [
         -- Using the fair conjunction operator (>>-) the test produces values
 
@@ -347,9 +355,9 @@ main = defaultMain $
         (nonTerminating $
          observeManyT 4 (let oddsPlus n = odds >>= \a -> return (a + n) in
                            (return 0 `mplus` return 1) >>-
-                           \a -> oddsPlus a >>=
+                           (oddsPlus >=>
                                  (\x -> if even x then return x else mzero)
-                        ))
+                        )))
 
         -- unfair conjunction does not terminate or produce any
         -- values: this will fail (expectedly) due to a timeout
@@ -370,7 +378,7 @@ main = defaultMain $
         )
 
       , testCase "fair conjunction :: ReaderT" $ [2,4,6,8] @=?
-        (take 4 $ runReaderT (let oddsR = return (1 :: Integer) `mplus` liftM (2+) oddsR
+        (take 4 $ runReaderT (let oddsR = return (1 :: Integer) `mplus` fmap (2+) oddsR
                                   oddsPlus n = oddsR >>= \a -> return (a + n)
                               in do x <- (return 0 `mplus` return 1) >>- oddsPlus
                                     if even x then return x else mzero
@@ -380,7 +388,7 @@ main = defaultMain $
       , testCase "fair conjunction :: CPS WriterT" $ [2,4,6,8] @=?
         (take 4 $ evalWriterT $
          (let oddsW :: CpsW.WriterT [Char] [] Integer
-              oddsW = return (1 :: Integer) `mplus` liftM (2+) oddsW
+              oddsW = return (1 :: Integer) `mplus` fmap (2+) oddsW
               oddsPlus n = oddsW >>= \a -> return (a + n)
            in do x <- (return 0 `mplus` return 1) >>- oddsPlus
                  if even x then return x else mzero
@@ -388,14 +396,14 @@ main = defaultMain $
 #endif
 
       , testCase "fair conjunction :: strict StateT" $ [2,4,6,8] @=?
-        (take 4 $ SS.evalStateT (let oddsS = return (1 :: Integer) `mplus` liftM (2+) oddsS
+        (take 4 $ SS.evalStateT (let oddsS = return (1 :: Integer) `mplus` fmap (2+) oddsS
                                      oddsPlus n = oddsS >>= \a -> return (a + n)
                                  in do x <- (return 0 `mplus` return 1) >>- oddsPlus
                                        if even x then return x else mzero
                                 ) "state")
 
       , testCase "fair conjunction :: lazy StateT" $ [2,4,6,8] @=?
-        (take 4 $ SL.evalStateT (let oddsS = return (1 :: Integer) `mplus` liftM (2+) oddsS
+        (take 4 $ SL.evalStateT (let oddsS = return (1 :: Integer) `mplus` fmap (2+) oddsS
                                      oddsPlus n = oddsS >>= \a -> return (a + n)
                                  in do x <- (return 0 `mplus` return 1) >>- oddsPlus
                                        if even x then return x else mzero
@@ -451,7 +459,7 @@ main = defaultMain $
           oddsL = [ 1 :: Integer ] `mplus` [ o | o <- [3..], odd o ]
           oc = [ n
                | n <- oddsL
-               , (n > 1)
+               , n > 1
                ] >>= \n -> ifte (do d <- iota (n - 1)
                                     guard (d > 1 && n `mod` d == 0))
                            (const mzero)
@@ -460,7 +468,7 @@ main = defaultMain $
          take 10 oc
 
     , let iota n = msum (map return [1..n])
-          oddsR = return (1 :: Integer) `mplus` liftM (2+) oddsR
+          oddsR = return (1 :: Integer) `mplus` fmap (2+) oddsR
           oc = do n <- oddsR
                   guard (n > 1)
                   ifte (do d <- iota (n - 1)
@@ -472,7 +480,7 @@ main = defaultMain $
 
 #if MIN_VERSION_mtl(2,3,0)
     , let iota n = msum (map return [1..n])
-          oddsW = return (1 :: Integer) `mplus` liftM (2+) oddsW
+          oddsW = return (1 :: Integer) `mplus` fmap (2+) oddsW
           oc :: CpsW.WriterT [Char] [] Integer
           oc = do n <- oddsW
                   guard (n > 1)
@@ -485,7 +493,7 @@ main = defaultMain $
 #endif
 
     , let iota n = msum (map return [1..n])
-          oddsS = return (1 :: Integer) `mplus` liftM (2+) oddsS
+          oddsS = return (1 :: Integer) `mplus` fmap (2+) oddsS
           oc = do n <- oddsS
                   guard (n > 1)
                   ifte (do d <- iota (n - 1)
@@ -496,7 +504,7 @@ main = defaultMain $
          (take 10 $ SS.evalStateT oc "state")
 
     , let iota n = msum (map return [1..n])
-          oddsS = return (1 :: Integer) `mplus` liftM (2+) oddsS
+          oddsS = return (1 :: Integer) `mplus` fmap (2+) oddsS
           oc = do n <- oddsS
                   guard (n > 1)
                   ifte (do d <- iota (n - 1)
@@ -578,7 +586,11 @@ main = defaultMain $
   ]
 
 safely :: IO Integer -> IO (Either String Integer)
-safely o = fmap (left (head . lines . show)) (try o :: IO (Either SomeException Integer))
+safely o = do
+  p <- try o
+  pure $ case p of
+    Left (err :: SomeException) -> Left $ maybe "" fst $ uncons $ lines $ show err
+    Right n -> Right n
 
 -- | This is used to test logic operations that don't typically
 -- terminate by running a parallel race between the operation and a
